<?php
require_once('../vendor/autoload.php'); // Adjust the path as needed
include('../settings/db/db.php');  // Database connection details

// Fetch invoice number from URL with input validation
$invoice = filter_input(INPUT_GET, 'invoiceNo', FILTER_SANITIZE_STRING);

// Prepare and execute the query with error handling
$result = $dbo->prepare("SELECT * FROM sales_list WHERE invoiceNo = :invoiceNo");
$result->bindParam(':invoiceNo', $invoice);

if (!$result->execute()) {
    die("Error fetching sales data: " . implode(", ", $result->errorInfo()));
}

$saleData = $result->fetch(PDO::FETCH_ASSOC);

if (!$saleData) {
    die("No sale data found for the specified invoice number.");
}

$entrant = $saleData['entrant'];

// Initialize customer variables
$customer = null;

// Check if customer_id exists in saleData
if (isset($saleData['customer_id'])) {
    $customer_id = $saleData['customer_id'];

    // Fetch customer details if customer_id exists
    $customerResult = $dbo->prepare("SELECT * FROM customer WHERE id = :id");
    $customerResult->bindParam(':id', $customer_id);
    $customerResult->execute();
    $customer = $customerResult->fetch(PDO::FETCH_ASSOC);
}

// Create new PDF document
$pdf = new TCPDF('P', 'mm', 'A4', true, 'UTF-8', false);
$pdf->SetCreator(PDF_CREATOR);
$pdf->SetAuthor('Your Company Name');
$pdf->SetTitle('Invoice');

// Set margins
$pdf->SetMargins(10, 10, 10); // Left, Top, Right
$pdf->SetAutoPageBreak(TRUE, 10); // Auto page break with bottom margin
// Disable header and footer
$pdf->setPrintHeader(false);
$pdf->setPrintFooter(false);

$pdf->AddPage();

// Add the first row (logo, company info, invoice info)
$pdf->SetFont('times', '', 12); // Set font for the content

// Define the width of each column for the first row
$col1Width = 50; // First column (logo)
$col2Width = 80; // Second column (company info)
$col3Width = 60; // Third column (invoice info)

// First column: Insert the logo
$logoPath = '../settings/uploads/' . $logo; // Adjust to the correct path
$pdf->Image($logoPath, 10, 10, 30, 30, '', '', '', true, 300, '', false, false, 0, false, false, false); // Adjust the logo size and position

// Second column: Company Information
$companyInfo = "Your Company Name\n1234 Street Name\nCity, State, ZIP\nPhone: (123) 456-7890\nEmail: info@yourcompany.com";
$pdf->SetXY(10 + $col1Width, 10); // Move to the right of the logo
$pdf->MultiCell($col2Width, 10, $companyInfo, 0, 'L', 0, 0); // Display company info

// Third column: Invoice Information
$invoiceInfo = "Receipt No: " . htmlspecialchars($saleData['invoiceNo']) . "\n" .
               "Date: " . date("F j, Y, g:i a") . "\n" .
               "Prepared By: " . htmlspecialchars($entrant);
$pdf->SetXY(10 + $col1Width + $col2Width, 10); // Move to the right of the company info
$pdf->MultiCell($col3Width, 10, $invoiceInfo, 0, 'L', 0, 1); // Display invoice info, move to the next line after

// Add some space between rows
$pdf->Ln(15);

// Second row: Customer Information
$customerInfo = $customer ? 
    "Customer Name: " . htmlspecialchars($customer['name']) . "\n" .
    "Address: " . htmlspecialchars($customer['address']) . "\n" .
    "Phone: " . htmlspecialchars($customer['phone']) . "\n" .
    "Email: " . htmlspecialchars($customer['email']) 
    : "Customer Name: Not Available";
$pdf->MultiCell(0, 10, $customerInfo, 0, 'L', 0, 1); // Full width for customer info

// Add some space after customer info
$pdf->Ln(10);

// Define total available width for the table (A4 width - left margin - right margin)
$availableWidth = 210 - 10 - 10; // 210mm (A4 width) - 10mm (left margin) - 10mm (right margin)

// Product table content
$pdf->SetFont('times', '', 12);
$n = 1;
$totalAmount = 0; // Initialize total amount variable

$salesResult = $dbo->prepare("SELECT trade_name, sell_price, (sell_price * qty_sold) AS amount, qty_sold FROM sales_list WHERE invoiceNo = :invoiceNo");
$salesResult->bindParam(':invoiceNo', $invoice);
$salesResult->execute();

while ($row = $salesResult->fetch(PDO::FETCH_ASSOC)) {
    $pdf->Cell(15, 8, $n++, 1); // Adjusted line height to 8
    $pdf->Cell($availableWidth - 90, 8, htmlspecialchars($row['trade_name']), 1); // Product Name column
    $pdf->Cell(25, 8, htmlspecialchars($row['qty_sold']), 1);
    $pdf->Cell(25, 8, number_format($row['sell_price'], 2), 1);
    $pdf->Cell(25, 8, number_format($row['amount'], 2), 1);
    $pdf->Ln();

    // Add to total amount
    $totalAmount += $row['amount']; // Update total amount
}

// Total Amount - In the second-to-last and last columns
$pdf->SetFont('times', 'B', 12);
$pdf->Cell(15, 8, '', 0); // Empty cell for alignment
$pdf->Cell($availableWidth - 90, 8, '', 0); // Empty cell for Product Name column
$pdf->Cell(25, 8, '', 0); // Empty cell for Qty column
$pdf->Cell(25, 8, 'Total:', 1);
$pdf->Cell(25, 8, number_format($totalAmount, 2), 1, 1, 'C'); // Center the total amount in the last column

// Add a newline before the barcode
$pdf->Ln(10);

// Barcode Part
use BarcodeBakery\Barcode\BCGcode128;
use BarcodeBakery\Common\BCGColor;
use BarcodeBakery\Common\BCGDrawing;

$colorBlack = new BCGColor(0, 0, 0);
$colorWhite = new BCGColor(255, 255, 255);

$code = new BCGcode128();
$code->setScale(2);
$code->setThickness(30);
$code->setForegroundColor($colorBlack);
$code->setBackgroundColor($colorWhite);
$code->setFont(null); // No font for no label
$code->setStart(null);
$code->setTilde(true);
//$code->parse($invoice66666666);
$code->parse($invoice . '666666');

// Drawing Part for barcode
$drawing = new BCGDrawing($code, $colorWhite);
ob_start(); // Start output buffering
$drawing->finish(BCGDrawing::IMG_FORMAT_PNG);
$barcodeImage = ob_get_contents(); // Get the image data
ob_end_clean(); // End output buffering

// Add the barcode image to the PDF
$pdf->Image('@' . $barcodeImage, '', '', 0, 18, '', '', '', true, 300, '', false, false, 0, false, false, false);

// Output the PDF
$pdf->Output('invoice_' . $invoice . '.pdf', 'I');
?>
